/**
 * @flow
 * @module configFactory
 */
/* eslint-disable no-unused-vars */

import path from 'path'
import webpack from 'webpack'
import StatsPlugin from 'stats-webpack-plugin'
import AssetsPlugin from 'assets-webpack-plugin'
import nodeExternals from 'webpack-node-externals'
import DuplicateCheckerPlugin from 'duplicate-package-checker-webpack-plugin'
import HardSourceWebpackPlugin from 'hard-source-webpack-plugin'
import { BundleAnalyzerPlugin } from 'webpack-bundle-analyzer'
import CaseSensitivePathsPlugin from 'case-sensitive-paths-webpack-plugin'
// import InterpolateHtmlPlugin from 'react-dev-utils/InterpolateHtmlPlugin'
import WatchMissingNodeModulesPlugin from 'react-dev-utils/WatchMissingNodeModulesPlugin'
import eslintFormatter from 'react-dev-utils/eslintFormatter'
// import ModuleScopePlugin from 'react-dev-utils/ModuleScopePlugin'
import ora from 'ora'
import ChunkManifestPlugin from 'chunk-manifest-webpack-plugin'
// import LodashWebpackPlugin from 'lodash-webpack-plugin'
import configPkg from '../../package.json'
// Custom plugins
import WebpackDigestHash from './plugins/ChunkHash'
import WebpackChunkNames from './plugins/ChunkNames'
import VerboseProgress from './plugins/VerboseProgress'

// Utility functions
import { removeEmpty, removeEmptyKeys, ifElse, isLoaderSpecificFile } from './helpers'

// Webpack blocks
import esModules from './modules/esModules'
import getJsLoader from './modules/getJsLoader'
import getCssLoader, { modulesPlugin, globalsPlugin } from './modules/getCssLoader'

/* eslint-disable-next-line */
// const DashboardPlugin = require('webpack-dashboard/plugin')
// Built in modules to exclude from bundle
const builtInSet = new Set(builtinModules)

// - "mime-db" database for working with mime types. Naturally pretty large stuff.
// - "helmet" uses some look with require which are not solvable with webpack
// - "express" also uses some dynamic requires
// - "encoding" uses dynamic iconv loading
// - "node-pre-gyp" native code module helper
// - "iltorb" brotli compression wrapper for NodeJS
// - "node-zopfli" native Zopfli implementation
const problematicCommonJS = new Set([
  'mime-db',
  'helmet',
  'express',
  'encoding',
  'node-pre-gyp',
  'iltorb',
  'node-zopfli',
  'import-inspector',
])

const cache = {
  'web-production': {},
  'web-development': {},
  'node-production': {},
  'node-development': {},
}

type Target = 'web' | 'node'
type Mode = 'development' | 'production'
type Config = {
  target: Target,
  mode: Mode,
  root?: string,
  isDebug?: boolean,
  isVerbose?: boolean,
  entry: Object,
  output?: Object,
  settings: any,
}

// eslint-disable-next-line max-statements, complexity
/**
 * Function which generates webpack config from passed options
 * @param {Config} options Build settings, target and stuff
 */
export default function configFactory(options: Config): Object {
  const {
    target,
    mode,
    output,
    root = process.cwd(),
    isDebug = false,
    isVerbose = false,
    entry = {},
    settings,
  } = options
  console.info('\n\n\n', `Build config version ${configPkg.version} targetting ${target}-${mode}`, '\n\n\n')
  // Should we use HardSource plugin
  const enableHardSource = !settings.disableHardsource

  if (!target || !['web', 'node'].includes(target)) {
    throw new Error('You must provide a "target" (web|node) to the configFactory.')
  }
  if (!mode || !['development', 'production'].includes(mode)) {
    throw new Error('You must provide a "mode" (development|production) to the configFactory.')
  }
  if (settings.clientBundleOutputPath === undefined) {
    throw new Error("No .frackrc.js file found. Please provide one to your project's root. ")
  }

  ora(`Creating bundle for ${target} in ${mode} mode.`).stopAndPersist('🤘')
  settings.nodeEnv = isDebug ? 'development' : mode
  settings.babelEnv = mode
  const version = JSON.stringify(require(path.join(root, 'package.json')).version)
  const isDev: boolean = mode === 'development'
  const isProd: boolean = mode === 'production'
  const isWeb: boolean = target === 'web'
  const isNode: boolean = target === 'node'
  const ifDev: Function = ifElse(isDev)
  const ifProd: Function = ifElse(isProd)
  const ifWeb: Function = ifElse(isWeb)
  const ifNode: Function = ifElse(isNode)
  const ifDevWeb: Function = ifElse(isDev && isWeb)
  const ifProdWeb: Function = ifElse(isProd && isWeb)
  const folder: string = isWeb ? 'client' : 'server'
  const CONFIG: Object = {
    isProd,
    isDev,
    isWeb,
    isNode,
    root,
  }
  const cssLoadersWithoutModules: Object = getCssLoader({ ...CONFIG, enableModules: false })
  const cssLoadersWithModules: Object = getCssLoader({ ...CONFIG, enableModules: true })
  const jsLoaders: Object | Array<Object> = getJsLoader(CONFIG)
  const excludeFromTranspilation = [
    /node_modules/,
    path.resolve(root, settings.clientBundleOutputPath),
    path.resolve(root, settings.serverBundleOutputPath),
  ]
  return {
    // We need to inform Webpack about our build target
    target: ifNode('node', 'web'),
    // We have to set this to be able to use these items when executing a
    // node bundle. Otherwise strangeness happens, like __dirname resolving
    // to '/'. There is no effect on our client bundle.
    node: {
      __dirname: true,
      __filename: true,
      process: true,
    },
    // What information should be printed to the console
    stats:
      isDebug || isVerbose
        ? {
            colors: true,
            reasons: isDebug,
            hash: isVerbose,
            version: isVerbose,
            timings: true,
            chunks: isVerbose,
            chunkModules: isVerbose,
            cached: isVerbose,
            cachedAssets: isVerbose,
          }
        : {
            reasons: true,
            colors: true,
          },
    // This is not the file cache, but the runtime cache.
    // The reference is used to speed-up rebuilds in one execution e.g. via watcher
    // Note: But is has to share the same configuration as the cache is not config aware.
    cache: cache[`${target}-${mode}`],
    // Capture timing information for each module.
    // Analyse tool: http://webpack.github.io/analyse
    profile: isProd,
    // Report the first error as a hard error instead of tolerating it.
    bail: isProd,
    externals: nodeExternals(),
    // See also: https://webpack.github.io/docs/configuration.html#devtool
    // and http://webpack.github.io/docs/build-performance.html#sourcemaps
    // All "module*" and "cheap" variants do not seem to work with this kind
    // of setup where we have loaders involved. Even simple console messages jump
    // to the wrong location in these cases.
    devtool: 'source-map',

    // New performance hints. Only active for production build.
    performance: isProd && isWeb ? { hints: 'warning' } : false,
    // Define our entry chunks for our bundle.
    entry:
      Object.keys(entry).length > 0
        ? entry
        : removeEmptyKeys({
            main: removeEmpty([
              'babel-polyfill',
              ifDevWeb(
                `webpack-hot-middleware/client?overlay=true&path=http://localhost:${
                  settings.clientDevserverPort
                }/__webpack_hmr`
              ),
              `./src/${folder}/index.js`,
            ]),
            config: ifProdWeb('./src/application/config.js'),
          }),
    output: {
      // The dir in which our bundle should be output.
      path: path.resolve(root, isWeb ? settings.clientBundleOutputPath : settings.serverBundleOutputPath),
      // The filename format for our bundle's entries.
      filename: ifProdWeb(
        // We include a hash for client caching purposes. Including a unique
        // has for every build will ensure browsers always fetch our newest bundle.
        `[name]-[chunkhash]-${version.replace(/[^0-9]/g, '')}.js`,
        // We want a determinable file name when running our NodeJS bundles,
        // as we need to be able to target our NodeJS start file from our
        // npm scripts. We don't care about caching on the NodeJS anyway.
        // We also want our client development builds to have a determinable
        // name for our hot reloading client bundle NodeJS.
        '[name].js'
      ),
      chunkFilename: ifProdWeb('[name]-[chunkhash].js', '[name].js'),
      // This is the web path under which our webpack bundled output should
      // be considered as being served from.
      publicPath: ifDev(
        // As we run a seperate NodeJS for our client and NodeJS bundles we
        // need to use an absolute http path for our assets public path.
        `http://localhost:${settings.clientDevserverPort}${settings.clientBundleHttpPath}`,
        // Otherwise we expect our bundled output to be served from this path.
        settings.clientBundleHttpPath
      ),
      // When in NodeJS mode we will output our bundle as a commonjs2 module.
      libraryTarget: ifNode('commonjs2', 'var'),
      // Override settings which has user provided in options
      ...output,
    },
    recordsOutputPath: path.resolve(settings.clientBundleOutputPath, 'records.json'),
    resolve: {
      // Enable new module/jsnext:main field for requiring files
      // Defaults: https://webpack.github.io/docs/configuration.html#resolve-packagemains
      mainFields: ifNode(
        enableHardSource ? ['main'] : ['module', 'jsnext:main', 'main'],
        enableHardSource
          ? ['web', 'browser', 'style', 'main']
          : ['web', 'browser', 'style', 'module', 'jsnext:main', 'main']
      ),
      modules: [
        path.resolve(root, 'node_modules'),
        path.resolve(root, 'src/application'),
        path.resolve(root, 'public'),
      ],
      alias: {
        'global-styles': path.resolve(root, 'src/application/global-styles'),
        application: path.resolve(root, 'src/application'),
        utils: path.resolve(root, 'src/application/utils'),
        store: path.resolve(root, 'src/application/store'),
        client: path.resolve(root, 'src/client'),
        server: path.resolve(root, 'src/server'),
        components: path.resolve(root, 'src/application/components'),
        modules: path.resolve(root, 'src/application/modules'),
        routes: path.resolve(root, 'src/application/routes'),
        services: path.resolve(root, 'src/application/services'),
        queries: path.resolve(root, 'src/application/queries'),
        public: path.resolve(root, 'public'),
      },
      // These extensions are tried when resolving a file.
      extensions: ['.js', '.jsx', '.json'],
    },
    plugins: removeEmpty([
      // Enforces the entire path of all required modules match the exact case of the actual path on disk.
      // https://github.com/Urthen/case-sensitive-paths-webpack-plugin
      new CaseSensitivePathsPlugin(),
      // new LodashWebpackPlugin(),
      // Improve source caching in Webpack v3.
      // Therefor we disable it in production and only use it to speed up development rebuilds.
      ifDev(
        new HardSourceWebpackPlugin({
          // // Either an absolute path or relative to output.path.
          // cacheDirectory: path.resolve(root, '.hardsource', `${target}-${mode}`),
          // // Either an absolute path or relative to output.path. Sets webpack's // recordsPath if not already set.
          // recordsPath: path.resolve(root, '.hardsource', `${target}-${mode}`, 'records.json'),
          // // Optional field. This field determines when to throw away the whole
          // // cache if for example npm modules were updated.
          // environmentHash: {
          //   root,
          //   directories: ['node_modules'],
          //   files: ['package.json', 'package-lock.json', 'yarn.lock', '.frackrc.js'],
          // },
        })
      ), // There is now actual benefit in using multiple chunks for possibly long living // NodeJS applications. We can bundle everything and that way improve startup time. // ifNode(new webpack.optimize.LimitChunkCountPlugin({ maxChunks: 1 })), // Adds options to all of our loaders.
      ifDev(
        new webpack.LoaderOptionsPlugin({
          // Indicates to our loaders that they should minify their output
          // if they have the capability to do so.
          minimize: false, // Indicates to our loaders that they should enter into debug mode // should they support it.
          debug: isDebug,
        })
      ), // Adds options to all of our loaders.
      ifProd(
        new webpack.LoaderOptionsPlugin({
          // Indicates to our loaders that they should minify their output
          // if they have the capability to do so.
          minimize: true, // Indicates to our loaders that they should enter into debug mode // should they support it.
          debug: isDebug,
        })
      ),
      // Verbose progress if verbose is enabled
      new VerboseProgress(),
      // For NodeJS bundle, you also want to use "source-map-support" which automatically sourcemaps
      // stack traces from NodeJS. We need to install it at the top of the generated file, and we
      // can use the BannerPlugin to do this.
      // - `raw`: true tells webpack to prepend the text as it is, not wrapping it in a comment.
      // - `entryOnly`: false adds the text to all generated files, which you might have multiple if using code splitting.
      // Via: http://jlongster.com/Backend-Apps-with-Webpack--Part-I
      ifNode(
        new webpack.BannerPlugin({
          banner: 'require("source-map-support").install();',
          raw: true,
          entryOnly: false,
        })
      ),
      // Extract vendor bundle for keeping larger parts of the application code
      // delivered to users stable during development (improves positive cache hits)
      ifProdWeb(
        new webpack.optimize.CommonsChunkPlugin({
          name: 'vendor',
          minChunks(module) {
            // this assumes your vendor imports exist in the node_modules directory
            return module.context && module.context.indexOf('node_modules') !== -1
          },
        })
      ),
      // Additional helper for non mentioned vendor like common modules
      // Docs: Look through all normal [aka lazy loaded] chunks, and if you find the same module that
      // appears across 3 or more chunks, then separate it out into a separate async commons chunk.
      // Via: https://medium.com/webpack/webpack-bits-getting-the-most-out-of-the-commonschunkplugin-ab389e5f318
      // ifProdWeb(new webpack.optimize.CommonsChunkPlugin({
      //   async: true,
      //   children: true,
      // })),
      // More aggressive chunk merging strategy. Even similar chunks are merged if the
      // total size is reduced enough.
      ifProdWeb(new webpack.optimize.AggressiveMergingPlugin()),
      //
      // We use this so that our generated [chunkhash]'s are only different if
      // the content for our respective chunks have changed.  This optimise
      // our long term browser caching strategy for our client bundle, avoiding
      // cases where browsers end up having to download all the client chunks
      // even though 1 or 2 may have only changed.
      ifProdWeb(new WebpackDigestHash()),
      // Extract chunk hashes into separate file
      ifProdWeb(
        new ChunkManifestPlugin({
          filename: 'manifest.json',
          manifestVariable: 'CHUNK_MANIFEST',
        })
      ),
      // Split to more chunks to optimize traffic on HTTP2
      // See also: https://medium.com/webpack/webpack-http-2-7083ec3f3ce6#.d66ve28x0
      // ifProdWeb(new webpack.optimize.AggressiveSplittingPlugin({
      //   minSize: 50000,
      //   maxSize: 120000,
      // })),
      ifProdWeb(new WebpackChunkNames()),
      new webpack.ContextReplacementPlugin(/moment[/\\]locale$/, /cs/),
      // Each key passed into DefinePlugin is an identifier.
      // The values for each key will be inlined into the code replacing any
      // instances of the keys that are found.
      // If the value is a string it will be used as a code fragment.
      // If the value isn’t a string, it will be stringified (including functions).
      // If the value is an object all keys are removeEmpty the same way.
      // If you prefix typeof to the key, it’s only removeEmpty for typeof calls.
      new webpack.DefinePlugin({
        // NOTE: The NODE_ENV key is especially important for production
        // builds as React relies on settings.nodeEnv for optimizations.
        'process.env.APP_ENV': JSON.stringify(process.env.APP_ENV),
        'process.env.NODE_ENV': JSON.stringify(process.env.NODE_ENV),
        // All the below items match the config items in our .env file. Go
        // to the .env.example for a description of each key.
        'process.env.API_ENDPOINT_REST_BASE': JSON.stringify(settings.apiEndpointRestBase),
        'process.env.BASE_URL': JSON.stringify(settings.baseUrl),
        'process.env.USE_REST_PROXY': JSON.stringify(settings.useRestProxy),
        'process.env.IS_BROWSER': isWeb,
        'process.env.VERSION': version,
        'process.env.TARGET': JSON.stringify(target),
        'process.env.MODE': JSON.stringify(mode),
        'process.env.SERVER_PORT': settings.serverPort,
        'process.env.CLIENT_DEVSERVER_PORT': settings.clientDevserverPort,
        'process.env.DISABLE_SSR': settings.disableSsr,
        'process.env.SERVER_BUNDLE_OUTPUT_PATH': JSON.stringify(settings.serverBundleOutputPath),
        'process.env.CLIENT_BUNDLE_OUTPUT_PATH': JSON.stringify(settings.clientBundleOutputPath),
        'process.env.CLIENT_PUBLIC_PATH': JSON.stringify(settings.clientPublicPath),
        'process.env.CLIENT_BUNDLE_ASSETS_FILENAME': JSON.stringify(settings.clientBundleAssetsFilename),
        'process.env.CLIENT_BUNDLE_CHUNK_MANIFEST_FILENAME': JSON.stringify(settings.clientBundleChunkManifestFilename),
        'process.env.CLIENT_BUNDLE_HTTP_PATH': JSON.stringify(settings.clientBundleHttpPath),
        'process.env.CLIENT_BUNDLE_CACHE_MAXAGE': JSON.stringify(settings.clientBundleCacheMaxage),
      }),
      // Generates a JSON file containing a map of all the output files for
      // our webpack bundle. A necessity for our NodeJS rendering process
      // as we need to interrogate these files in order to know what JS/CSS
      // we need to inject into our HTML.
      ifWeb(
        new AssetsPlugin({
          filename: 'assets.json',
          path: path.join(root, settings.clientBundleOutputPath),
          prettyPrint: true,
          update: false,
        })
      ),
      ifProdWeb(new StatsPlugin('stats.json')),
      // We don't want webpack errors to occur during development as it will
      // kill our development NodeJS instances.
      ifDev(new webpack.NoEmitOnErrorsPlugin()),
      // We need this plugin to enable hot module reloading for our dev NodeJS.
      ifDevWeb(new webpack.HotModuleReplacementPlugin()),
      // This is a production client so we will extract our CSS into // CSS files.
      ifProdWeb(modulesPlugin),
      ifProdWeb(globalsPlugin),
      // Warns you when multiple versions of the same package exist in a build.
      // https://github.com/darrenscerri/duplicate-package-checker-webpack-plugin
      ifProdWeb(new DuplicateCheckerPlugin()),
      // Analyse webpack bundle
      ifProdWeb(
        new BundleAnalyzerPlugin({
          openAnalyzer: false,
          analyzerMode: 'static',
          logLevel: 'error',
        })
      ),
      // Minify assets for production web compilation
      ifProdWeb(
        new webpack.optimize.UglifyJsPlugin({
          compress: { drop_console: true },
          sourceMap: true,
          mangle: true,
          warnings: false,
          comments: false,
          parallel: { cache: true, workers: 4 },
        })
      ),
      // Provide global plugins
      new webpack.ProvidePlugin({
        Promise: 'bluebird',
      }),
      new WatchMissingNodeModulesPlugin(path.resolve(process.cwd(), 'node_modules')),
    ]),
    module: {
      rules: removeEmpty([
        // {
        //   test: /\.(js|jsx)$/,
        //   enforce: 'pre',
        //   use: [
        //     {
        //       options: {
        //         formatter: eslintFormatter,
        //         eslintPath: require.resolve('eslint'),
        //         baseConfig: {
        //           extends: [require.resolve('eslint-config-frack')],
        //         },
        //         ignore: false,
        //         useEslintrc: true,
        //       },
        //       loader: require.resolve('eslint-loader'),
        //     },
        //   ],
        //   include: [path.resolve(process.cwd(), `/src/${folder}/`)],
        //   exclude: excludeFromTranspilation,
        // },
        {
          test: /\.(svg)$/,
          include: [/src/, /assets/],
          loader: 'babel-loader?cacheDirectory=true,presets[]=es2015,presets[]=react!svg-react-loader',
        },
        {
          // JavaScript
          test: /\.(mjs|js|jsx)$/,
          loaders: jsLoaders,
          exclude: excludeFromTranspilation,
        },
        {
          // Markdown Container MDX Format
          // See also: http://jamesknelson.com/introducing-mdxc/
          test: /\.mdx?$/,
          use: jsLoaders.concat(['mdx-loader']),
        },
        {
          // CSS modules
          test: /[A-Z][A-z0-9]+\.(scss|css|sss|pcss)$/,
          loader: cssLoadersWithModules,
          include: [/src/],
          exclude: [/node_modules/, /global-styles/],
        },
        {
          // Plain CSS
          test: /[a-z][A-z0-9]+\.(scss|css|sss|pcss)$/,
          loader: cssLoadersWithoutModules,
          include: [/node_modules/, /src/, /global-styles/],
          exclude: [/components/],
        },
        {
          // JSON
          test: /\.json$/,
          loader: 'json-loader',
          exclude: [],
        },
        {
          // YAML
          test: /\.(yml|yaml)$/,
          loaders: ['json-loader', 'yaml-loader'],
        },
        {
          // References to images, fonts, movies, music, etc.
          test: /\.(eot|woff|woff2|ttf|otf|png|jpg|jpeg|jp2|jpx|jxr|gif|webp|mp4|mp3|ogg|pdf|html)$/,
          loader: 'file-loader',
          options: {
            name: ifProdWeb('file-[hash:base62:8].[ext]', '[name].[ext]'),
            emitFile: isWeb,
          },
        },
        {
          // GraphQL support
          // @see http://dev.apollodata.com/react/webpack.html
          test: /\.(graphql|gql)$/,
          loader: 'graphql-tag/loader',
          exclude: excludeFromTranspilation,
        },
      ]),
    },
  }
}
