// @flow

import ExtractTextPlugin from 'extract-text-webpack-plugin'

export const modulesPlugin = new ExtractTextPlugin({
  filename: 'modules-[contenthash:base62:8].css',
  allChunks: true,
})
export const globalsPlugin = new ExtractTextPlugin({
  filename: 'global-[contenthash:base62:8].css',
  allChunks: true,
})

/**
 * @interface CSSLoaderFactoryOptions
 */
type CSSLoaderFactoryOptions = {
  /**
   * Is build target NodeJS
   */
  isNode: boolean,
  /**
   * Is build target web
   */
  isWeb: boolean,
  /**
   * Is production build
   */
  isProd: boolean,
  /**
   * Enable CSS modules
   */
  enableModules: boolean,
}
/**
 * CSS Loader factory
 * @param {CSSLoaderFactoryOptions} options Configuration toggles
 */
export default function getCssLoaders(options: CSSLoaderFactoryOptions): any {
  const {
    isNode, isWeb, isProd, enableModules = true,
  } = options
  const moduleIdentName = enableModules ? '[name]__[local]' : '[local]'
  // When targetting the node we fake out the style loader as the
  // node can't handle the styles and doesn't care about them either..
  if (isNode) {
    return [
      {
        loader: 'css-loader/locals',
        query: {
          sourceMap: false,
          modules: enableModules,
          localIdentName: moduleIdentName,
          importLoaders: 2,
          minimize: false,
        },
      },
      {
        loader: 'postcss-loader',
        query: {
          sourceMap: false,
        },
      },
      {
        loader: 'fast-sass-loader',
        query: {
          sourceMap: false,
        },
      },
    ]
  }
  if (isWeb) {
    // For a production client build we use the ExtractTextPlugin which
    // will extract our CSS into CSS files. The plugin needs to be
    // registered within the plugins section too.
    if (isProd) {
      // First: the loader(s) that should be used when the css is not extracted
      // Second: the loader(s) that should be used for converting the resource to a css exporting module
      // Note: Unfortunately it seems like it does not support the new query syntax of webpack v2
      // See also: https://github.com/webpack/extract-text-webpack-plugin/issues/196
      return (enableModules ? modulesPlugin : globalsPlugin).extract({
        allChunks: true,
        fallback: 'style-loader',
        filename: 'css/[name]-[contenthash].css',
        use: [
          {
            loader: 'css-loader',
            query: {
              modules: enableModules,
              localIdentName: moduleIdentName,
              minimize: false,
              importLoaders: 2,
              sourceMap: false,
            },
          },
          {
            loader: 'postcss-loader',
            query: {
              sourceMap: false,
            },
          },
          {
            loader: 'fast-sass-loader',
            query: {
              sourceMap: false,
            },
          },
        ],
      })
    }
    // For a development client we will use a straight style & css loader
    // along with source maps. This combo gives us a better development
    // experience.
    return [
      {
        loader: 'style-loader',
      },
      {
        loader: 'css-loader',
        query: {
          sourceMap: false,
          modules: enableModules,
          localIdentName: moduleIdentName,
          minimize: false,
          importLoaders: 2,
        },
      },
      {
        loader: 'postcss-loader',
        query: {
          sourceMap: false,
        },
      },
      {
        loader: 'fast-sass-loader',
        query: {
          sourceMap: false,
        },
      },
    ]
  }
  return []
}
