// @flow

import deepMerge from 'deepmerge'
import { resolve } from 'path'

// Common babel configuration
const BABEL_COMMON = {
  // Don't try to find .babelrc because we want to force this configuration.
  babelrc: false,
  // Faster transpiling for minor loose in formatting
  compact: true,
  // Keep origin information alive
  sourceMap: true,
  // Nobody needs the original comments when having source maps
  comments: false,
  // Enable caching for babel transpiles
  cacheDirectory: true,

  presets: [
    // Decorators
    require.resolve('babel-preset-stage-0'),
    // Flow support
    require.resolve('babel-preset-flow'),
    // JSX, Flow
    require.resolve('babel-preset-react'),
  ],

  // Common plugins
  plugins: [
    // Use helpers, but not polyfills, in a way that omits duplication.
    [
      require.resolve('babel-plugin-macros'),
      require.resolve('babel-plugin-transform-runtime'),
      {
        regenerator: true,
        polyfill: true,
        useBuiltIns: true,
        useESModules: false,
      },
    ],
    // JSX-Control-Statements is a Babel plugin that extends JSX to add basic
    // control statements: conditionals and loops.
    require.resolve('babel-plugin-jsx-control-statements'),
    // flow syntax support for better errors
    // require.resolve('babel-plugin-syntax-flow'),
    // SSR support for React Loadable
    [
      require.resolve('babel-plugin-import-inspector'),
      {
        serverSideRequirePath: true,
        webpackRequireWeakId: true,
      },
    ],
    // function* () { yield 42; yield 43; }
    [
      require.resolve('babel-plugin-transform-regenerator'),
      {
        // Async functions are converted to generators by babel-preset-env
        async: false,
      },
    ],
    // Adds syntax support for import()
    require.resolve('babel-plugin-syntax-dynamic-import'),
    // class { handleClick = () => { } }
    // https://github.com/tc39/proposal-class-public-fields
    require.resolve('babel-plugin-transform-class-properties'),
    // { ...todo, completed: true }
    [require.resolve('babel-plugin-transform-object-rest-spread'), { useBuiltIns: true }],
  ],

  // Environment specific config
  env: {
    // Production specifics
    production: {
      // Remove comments
      comments: false,
      plugins: [
        // support jsx control statements
        require.resolve('babel-plugin-jsx-control-statements'),
        // Remove prop types from our code
        [require.resolve('babel-plugin-transform-react-remove-prop-types'), { removeImport: true }],
      ],
    }, // Development specifics
    development: {
      plugins: [
        //  React PropTypes definitions from Flow type declarations
        require.resolve('babel-plugin-jsx-control-statements'),
        // require.resolve('babel-plugin-transform-react-jsx-source'), // Adds __self attribute to JSX which React will use for some warnings
        // require.resolve('babel-plugin-transform-react-jsx-self'), // Add deprecation messages
        require.resolve('babel-plugin-log-deprecated'), // Warns about deprecated
      ],
    },
  },
}

// NodeJS specific config
const BABEL_NODE = {
  presets: [
    // Exponentiation
    [
      require('babel-preset-env').default,
      {
        targets: {
          node: 8.0,
          uglify: true,
        },
        // Do not transform modules to CJS
        modules: false,
      },
    ],
  ],
  plugins: [
    // Compiles import() to a deferred require()
    require.resolve('babel-plugin-dynamic-import-node'),
  ],
}

// Web specific config
const BABEL_WEB = {
  presets: [
    [
      require.resolve('babel-preset-env'),
      {
        targets: {
          safari: '9.1.3',
          // React parses on ie 9, so we should too
          ie: 11,
          // We currently minify with uglify
          // Remove after https://github.com/mishoo/UglifyJS2/issues/448
          uglify: true,
        },
        // Disable polyfill transforms
        useBuiltIns: false,
        // Do not transform modules to CJS
        modules: false,
      },
    ],
  ],
}

/**
 * Get Javascript Loader for passed config
 * @param config    {isProd: boolean, isDev: boolean, babelConfig: Object, root: string}    Configuration object
 * @return          {Object}     Webpack loader config
 */
export default function getJsLoader({
  isNode, isWeb, babelConfig = {}, root,
}) {
  const resolverPlugin = {
    plugins: [
      [
        'module-resolver',
        {
          root: [resolve(root, 'src')],
        },
      ],
    ],
  }
  const commonBabel = deepMerge(resolverPlugin, BABEL_COMMON)
  const nodeBabel = isNode ? BABEL_NODE : {}
  const webBabel = isWeb ? BABEL_WEB : {}
  const config = [
    {
      loader: 'babel-loader',
      options: deepMerge(deepMerge(nodeBabel, webBabel), deepMerge(babelConfig, commonBabel)),
    },
  ]

  return config
}
