// @flow

import { startsWith } from 'lodash'
import ora from 'ora'
import prettyMs from 'pretty-ms'

export default class VerboseProgress {
  options: any = null
  constructor(options: any) {
    this.options = options
  }

  apply(compiler: any) {
    let modules: typeof ora
    let seal: typeof ora
    let optimize: typeof ora
    let optimizeAssets: typeof ora
    let emit: typeof ora

    compiler.plugin('compilation', (compilation) => {
      if (compilation.compiler.isChild()) {
        return
      }

      let moduleCounter = 0

      compilation.plugin('build-module', (module) => {
        const ident = module.identifier()
        if (ident) {
          if (startsWith(ident, 'ignored') || startsWith(ident, 'external')) {
            return
          }
          if (moduleCounter === 0) {
            modules = ora({ text: 'Building module', spinner: 'clock' }).start()
          }
          moduleCounter += 1
          if (moduleCounter % 10 === 0) {
            modules.text = `Bundling module #${moduleCounter}`
          }
        }
      })

      compilation.plugin('seal', () => {
        modules && modules.succeed()
        seal = ora({ text: `Sealing ${moduleCounter} modules`, spinner: 'clock' })
      })

      compilation.plugin('optimize', () => {
        seal && seal.succeed()
        optimize = ora({ text: 'Optimizing modules', spinner: 'clock' }).start()
      })

      compilation.plugin('optimize-chunk-assets', (chunks, callback) => {
        optimize && optimize.succeed()
        optimizeAssets = ora({ text: 'Optimizing assets', spinner: 'clock' }).start()
        callback()
      })

      compilation.plugin('after-optimize-chunk-assets', (chunks, callback) => {
        optimizeAssets && optimizeAssets.succeed()
      })
    })

    compiler.plugin('emit', (compilation, callback) => {
      optimizeAssets.succeed()
      emit = ora({ text: 'Emitting output files', spinner: 'clock' }).start()
      callback()
    })

    compiler.plugin('after-emit', (compilation, callback) => {
      emit && emit.succeed()
      callback()
    })

    compiler.plugin('done', (stats) => {
      ora({
        text: `Compilation took ${prettyMs(stats.endTime - stats.startTime)}\n`,
      }).stopAndPersist('🤘')
    })
  }
}
